#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

Player *botInfo::getPlayer(int ident)
{
	_listnode <Player> *parse = playerlist->head;

	while (parse)
	{
		Player *item = parse->item;

		if (item->ident == ident)
			return item;

		parse = parse->next;
	}

	return NULL;
}

int botInfo::getFlagCount(int team)
{
	_listnode <Flag> *parse = flaglist->head;

	int count = 0;

	while (parse)
	{
		Flag *item = parse->item;

		if (item->team)
			++count;

		parse = parse->next;
	}

	return count;
}

struct TeamEntry
{
	int team, flags;

	TeamEntry(int t, int c)
	{
		team = t;
		flags = c;
	}

	bool validate(int t, int c)
	{
		if (t != team) return false;

		flags += c;

		return true;
	}
};

String botInfo::getFlags()
{
	String s;

	int seen = 0;

	_linkedlist <TeamEntry> teamlist;

	// compile team ownings

	_listnode <Player> *parse = playerlist->head;

	while (parse)
	{
		Player *p = parse->item;

		// check for prior entry

		if (p->flagCount)
		{
			_listnode <TeamEntry> *search = teamlist.head;

			while (search)
			{
				TeamEntry *e = search->item;

				if (e->validate(p->team, p->flagCount))
					break;

				search = search->next;
			}

			if (!search)
			{
				teamlist.append(new TeamEntry(p->team, p->flagCount));
			}
		}

		parse = parse->next;
	}

	// write results

	_listnode <TeamEntry> *search = teamlist.head;

	while (search)
	{
		TeamEntry *e = search->item;

		if (e->team != MAX_USERS)
		{
			if (seen) s += " ";

			if (e->team < MAX_PUBLIC_TEAMS)
			{
				s += "(";
				s += e->team;
				s += "):";
			}
			else
				s += "(Prv):";
			s += e->flags;

			++seen;
		}

		search = search->next;
	}

	return s;
}

inline int abs(int n)
{
	if (n < 0)	return -n;
	else		return n;
}

bool closeto(Player *p, int x, int y, int tolerance)
{
	return (abs((p->tile.x) - x) < tolerance) && (abs((p->tile.y) - y) < tolerance);
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (countdown[0] == 5)
				sendPublic("?jackpot");

			if (countdown[0] == 0)
			{
				if (jackpot)
				{
					String s;
					s += "*arena Jackpot: ";
					s += jackpot;
					s += "  ";
					s += getFlags();
					sendPublic(s);

					countdown[0] = 15 * 60;
				}
				else
				{
					countdown[0] = 5 * 60;
				}
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];

			int count = 0;

			String s;
			_listnode <Player> *parse = playerlist->head;

			String name = "Name", squad = "Squad", win = "Win", loss = "Loss", points = "Flag pts";

			while (parse)
			{
				Player *p = parse->item;

				if (p->team == team)
				{
					if (!count)
					{
						s = "*arena ";
						s += name.rfill(19) + squad.rfill(19) + win.rfill(6) + loss.rfill(6) + points.rfill(10);
						sendPublic(s);
					}

					name = p->name;
					squad = p->squad;
					win = p->score.wins;
					loss = p->score.losses;
					points = p->score.flagPoints;

					s = "*arena ";
					s += name.rfill(19) + squad.rfill(19) + win.rfill(6) + loss.rfill(6) + points.rfill(10);
					sendPublic(s);

					++count;
				}

				parse = parse->next;
			}
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];

			if (p->ship != SHIP_Spectator)
			if (p->awarp)
			if (closeto(p, 512, 512, radius))
			{
				sendPrivate(p, "*prize #-20");
			}
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
					if (CMPSTART(msg, "Current Jackpot: "))
						jackpot = getInteger(msg + 17, 10);
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];

			if (major != CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  Major DLL version mismatch!"));

				CONNECTION_DENIED = true;

				return;
			}
			else if (minor != CORE_MINOR_VERSION)
				tell(makeEcho("DLL plugin connected.  Minor DLL version mismatch!"));
			else
				tell(makeEcho("DLL plugin connected."));
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	default:
		{
			String s;
			s += "Unknown message from bot (";
			s += event.code;
			s += ")  Version mismatch?";
			tell(makeEcho(s));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
